from __future__ import division
import sys
import gobject
import gtk
import gtk.gdk as gdk
import pango
import bonobo
import gosd
import gconfsync
import gc
import os.path
import gnome
import gnome_osd_conf
import warnings
import xml.dom.minidom as dom
import gtasklet
import traceback

try:
    import dbus
    import dbus.service
    import dbus.glib
except ImportError:
    HAVE_DBUS = False
else:
    HAVE_DBUS = True


bgcolor = "#000000"
fgcolor = "#80ff80"
gosd.BORDER_WIDTH = 1
ANIMATION = True
ANIMATION_STEPS = 10
ANIMATION_TIME = 200

gnome.program_init("gnome-osd-server", gnome_osd_conf.VERSION)
prefs = gconfsync.GConfSync("/apps/gnome-osd",
                            dict(osd_halignment=dict(left=pango.ALIGN_LEFT,
                                                     right=pango.ALIGN_RIGHT,
                                                     center=pango.ALIGN_CENTER)))

# Note: interning atoms avoids X server roundtrips
_NET_CURRENT_DESKTOP = gtk.gdk.atom_intern("_NET_CURRENT_DESKTOP")
_NET_WORKAREA = gtk.gdk.atom_intern("_NET_WORKAREA")
def get_current_workarea():
    root = gtk.gdk.screen_get_default().get_root_window()
    prop = root.property_get(_NET_CURRENT_DESKTOP)
    if prop is None:
        return None
    propname, proptype, propvalue = prop
    current = propvalue[0]
    prop = root.property_get(_NET_WORKAREA)
    if prop is None:
        return None
    propname, proptype, propvalue = prop
    return propvalue[current*4 : (current + 1)*4]

def _map_ellipsize_val(val):
    return {
        'none': pango.ELLIPSIZE_NONE,
        'start': pango.ELLIPSIZE_START,
        'middle': pango.ELLIPSIZE_MIDDLE,
        'end': pango.ELLIPSIZE_END,
        } [val.lower()]

    
class OSDMessage(object):

    class Message(object):
        __slots__ = ('text', 'timeout')

    class Animation(gtasklet.Tasklet):
        def __init__(self, win, x0, y0, xf, yf):
            self.x = float(x0)
            self.y = float(y0)
            self.dx = (xf - x0) / ANIMATION_STEPS
            self.dy = (yf - y0) / ANIMATION_STEPS
            self.win = win
            self.counter = ANIMATION_STEPS
            self.timeout = gtasklet.WaitForTimeout(ANIMATION_TIME//ANIMATION_STEPS)
            win.move(int(x0), int(y0))
            win.show()
            gtasklet.Tasklet.__init__(self)

        def run(self):
            counter = ANIMATION_STEPS
            while counter:
                counter -= 1

                yield self.timeout
                gtasklet.get_event()

                self.x += self.dx
                self.y += self.dy
                self.win.move(int(self.x), int(self.y))

    
    def __init__(self, id, inherit=None):
        self.hide_timeout = None
        self.message_queue = []
        self.__show_lock = False
        self.animation_vector = None
        self.inherit = inherit
        self.props = {}
        self.id = id
        self.occupied_area = None
        self.__show_task = None

    def getprop(self, name):
        try:
            return self.props[name]
        except KeyError:
            if self.inherit is None:
                return prefs[name]
            else:
                return self.inherit.getprop(name)

    def show(self, text, timeout=None):
        msg = self.Message()
        msg.text = text
        if timeout is None:
            msg.timeout = self.getprop('hide_timeout')
        else:
            msg.timeout = timeout
        self.message_queue.append(msg)
        if self.__show_task is None:
            self.__show_task = gtasklet.Tasklet(self._show_messages())
        else:
            ## speed up hiding existing message
            self.__show_task.send_message(gtasklet.Message("continue"))

    def _show_messages(self):
        while self.message_queue:
            msg = self.message_queue.pop(0)
            halign = self.getprop('osd_halignment')
            vpos = self.getprop('osd_vposition')

            if self.getprop('avoid_panels'):
                workarea = get_current_workarea()
                if workarea is not None:
                    desk_x, desk_y, desk_w, desk_h = workarea
                else:
                    warnings.warn("get_current_workarea failed, somehow")
                    desk_x, desk_y = 0, 0
                    desk_w = gdk.screen_width()
                    desk_h = gdk.screen_height()
            else:
                desk_x, desk_y = 0, 0
                desk_w = gdk.screen_width()
                desk_h = gdk.screen_height()

            w = gosd.osd(msg.text, bgcolor, fgcolor,
                         pango.FontDescription(self.getprop('osd_font')),
                         use_markup=True,
                         alignment=halign,
                         fake_translucent_bg=self.getprop("osd_fake_translucent_bg"),
                         drop_shadow=self.getprop("drop_shadow"),
                         max_width=desk_w,
                         debug_frame=self.getprop('debug_frame'),
                         ellipsize=_map_ellipsize_val(self.getprop('ellipsize')))
            if self.getprop('enable_sound'):
                gnome.sound_play(self.__get_sound_sample_file())

            y, x, y_final = self.reserve_display_area(vpos, halign, w,
                                                      desk_x, desk_y, desk_w, desk_h)
            if self.getprop('animations'):
                anim = self.Animation(w, x, y, x, y_final)
                yield gtasklet.WaitForTasklet(anim), gtasklet.WaitForMessages(defer='continue')
                gtasklet.get_event()
            else:
                w.move(x, y_final)
                w.show()

            wait_list = [gtasklet.WaitForMessages(accept='continue')]
            
            if not self.message_queue:
                ## if there are messages waiting, don't pause
                wait_list.append(gtasklet.WaitForTimeout(msg.timeout))
            if self.getprop('hide_on_hover'):
                wait_list.append(gtasklet.WaitForSignal(w, "enter-notify-event"))

            yield wait_list ## during this yield the message stays on screen, very still
            event = gtasklet.get_event()

            ## Now hide it
            if self.getprop('animations'):
                anim = self.Animation(w, x, y_final, x, y)
                yield gtasklet.WaitForTasklet(anim), gtasklet.WaitForMessages(defer='continue')
                gtasklet.get_event()

            w.destroy()
            self.occupied_area = None
            while gc.collect():
                pass

        self.__show_task = None
        return
        
    def __get_sound_sample_file(self):
        sound_sample = self.getprop('sound_sample')
        if sound_sample[0] == '/':
            return sound_sample
        else:
            return os.path.join(gnome_osd_conf.datadir, "sounds", sound_sample)

    instances = {}

    def get(cls, attrs={}):
        global prefs
        id_ = attrs.get('id', 'default')
        inherit = attrs.get('inherit', 'default')
        try:
            osd_msg = cls.instances[id_]
        except KeyError:
            if inherit == 'default':
                inherit_osd = None
            else:
                inherit_osd = cls.instances[inherit]
            osd_msg = OSDMessage(id=id_, inherit=inherit_osd)
            cls.instances[id_] = osd_msg
        for key, str_val in attrs.iteritems():
            if key in ('id', 'inherit'): # special properties, no gconf schema
                continue
            if id_ == 'default':
                raise ValueError, "changing message properties"\
                      " for 'default' instance not allowed"
            val = prefs.parse_string(key, str_val)
            osd_msg.props[key] = val
        return osd_msg
    get = classmethod(get)

    def reserve_display_area(self, vpos, halign, w, desk_x, desk_y, desk_w, desk_h):
        if halign == pango.ALIGN_LEFT:
            x = 2 + desk_x
        elif halign == pango.ALIGN_RIGHT:
            x = desk_x + desk_w - w.width - 2
        else: # center
            x = desk_x + desk_w/2 - w.width/2

        occupied_areas = [inst.occupied_area for inst in self.instances.itervalues()
                          if not (inst is self or inst.occupied_area is None)]

        def intersection(area, x, y):
            x1, y1 = x, y
            x2, y2 = x + w.width, y + w.height
            ox1, oy1, ox2, oy2 = area
            result = ((x1 >= ox1 and x1 <= ox2
                       or x2 <= ox2 and x2 >= ox1
                       or x1 <= ox1 and x2 >= ox2)
                      and
                      (y1 >= oy1 and y1 <= oy2
                       or y2 <= oy2 and y2 >= oy1
                       or y1 <= oy1 and y2 >= oy2))
            return result
            
        if vpos == 'top':
            y = -w.height
            def sort_func(a, b):
                return cmp(a[3], b[3])
            occupied_areas.sort(sort_func)
            y_final = desk_y + 2
            for area in occupied_areas:
                if intersection(area, x, y_final):
                    y_final = area[3] + 2
                else:
                    break
        elif vpos == 'center':
            y = -w.height
            y_ideal = desk_y + desk_h//2  - w.height//2
            def sort_func(a, b):
                return cmp(abs(y_ideal - a), abs(y_ideal - b))
            y_final = y_ideal
            possible_targets = [y_ideal]
            for area in occupied_areas:
                possible_targets.append(area[3] + 2)
                possible_targets.append(area[1] - 2 - w.height)
            possible_targets.sort(sort_func)
            for target in possible_targets:
                for area in occupied_areas:
                    if intersection(area, x, target):
                        break
                else:
                    y_final = target
                    break
            else:
                y_final = None
            assert y_final is not None
        elif vpos == 'bottom':
            y = gdk.screen_height()
            y_final = desk_y + desk_h - w.height - 2
            def sort_func(a, b):
                return cmp(b[1], a[1])
            occupied_areas.sort(sort_func)
            for area in occupied_areas:
                if intersection(area, x, y_final):
                    y_final = area[1] - 2 - w.height
                else:
                    break
        else:
            raise ValueError, "invalid vposition '%s'" % vpos
        self.occupied_area = (x, y_final, x + w.width, y_final + w.height)
        return y, x, y_final
        
    
def message_quit_cb(app_client, message, app):
    print "received quit message"
    app.unref()
    bonobo.main_quit()

def message_show_cb(_, message, timeout=-1):
    try:
        attrs = {'id': 'default'}
        if timeout == -1:
            timeout = None
        osd_msg = OSDMessage.get(attrs)
        osd_msg.show(message, timeout)
        return ''
    except Exception, ex:
        exc_info = sys.exc_info()
        try:
            error = "".join(traceback.format_exception(*exc_info))
        finally:
            del exc_info
        print >> sys.stderr, error
        return error

def message_parse(message):
    '''message_parse(msg) -> (attrs, pango_markup)'''
    doc = dom.parseString(message.decode("utf-8"))
    root = doc.documentElement
    if root.tagName != "message":
        raise RuntimeError("Protocol error: expecting <message>, got <%s>" % root.tagName)
    pango_markup = "".join(map(dom.Node.toxml, root.childNodes))
    attrs = {}
    for i in xrange(root.attributes.length):
        name = root.attributes.item(i).name
        attrs[name] = root.getAttribute(name)
    return attrs, pango_markup

def message_show_full_cb(_, message):
    try:
        attrs, pango_s = message_parse(message)
        osd_msg = OSDMessage.get(attrs)
        osd_msg.show(pango_s)
        return ''
    except Exception, ex:
        print >> sys.stderr, ex
        return "%s: %s" % (str(ex.__class__), str(ex))

(ACTIVATION_REG_SUCCESS, 
 ACTIVATION_REG_NOT_LISTED, 
 ACTIVATION_REG_ALREADY_ACTIVE, 
 ACTIVATION_REG_ERROR) = range(4)


def start_bonobo_interface():
    os.putenv("DISPLAY", gtk.gdk.get_display())
    bonobo.activate()
    app = bonobo.Application("GNOME_OSD")
    def last_unref_cb(*args):
        print "Last Unref, quitting"
        bonobo.main_quit()
    reg = bonobo.activation.active_server_register("OAFIID:GNOME_OSD", app.corba_objref())
    if reg != ACTIVATION_REG_SUCCESS:
        raise SystemExit(3)
    app.connect("message::quit", message_quit_cb, app)

    app.register_message("show",
                         "Show a message in pango markup using default options",
                         str, (str, int),
                         message_show_cb)

    app.register_message("show-full",
                         "Show a message in Gnome OSD protocol format"
                         " (XML <message ...> ... </message>)",
                         str, (str,),
                         message_show_full_cb)

    bonobo.context_running_get().connect("last-unref", last_unref_cb)

if HAVE_DBUS:
    class DBusInterface(dbus.service.Object):
        def __init__(self, bus_name, object_path="/Server"):
            dbus.service.Object.__init__(self, bus_name, object_path)

        #@dbus.service.method("pt.inescporto.telecom.GnomeOSD")
        def showMessage(self, message, timeout=-1):
            return message_show_cb(None, message, timeout)
        showMessage = dbus.service.method("pt.inescporto.telecom.GnomeOSD")(showMessage)

        #@dbus.service.method("pt.inescporto.telecom.GnomeOSD")
        def showMessageFull(self, message):
            return message_show_full_cb(None, message)
        showMessageFull = dbus.service.method("pt.inescporto.telecom.GnomeOSD")(showMessageFull)

    def start_dbus_interface():
        session_bus = dbus.SessionBus()
        name = dbus.service.BusName("pt.inescporto.telecom.GnomeOSD",
                                    bus=session_bus)
        obj = DBusInterface(name)
    

def main():
    print "starting Bonobo interface..."
    start_bonobo_interface()
    if HAVE_DBUS:
        print "starting DBUS interface..."
        start_dbus_interface()
    print "entering main loop..."
    bonobo.main()



'''
Some code to monitor xscreensaver
'''

import gobject

class XScreenSaverMonitor(gobject.GObject):

    __gsignals__ = { 'state-changed': (gobject.SIGNAL_RUN_LAST, None,
                                       (str,)) }

    def __init__(self):
        '''Starts monitoring xscreensaver; raises GError if xscreensaver-command not available'''
        gobject.GObject.__init__(self)
        pid, _, stdout, _ = gobject.spawn_async(
            [ 'xscreensaver-command', '-watch'],
            standard_input=False, standard_output=True, standard_error=False,
            flags=gobject.SPAWN_SEARCH_PATH)
        chan = gobject.IOChannel(filedes=stdout)
        chan.set_flags(gobject.IO_FLAG_NONBLOCK)
        chan.add_watch(gobject.IO_IN, self.__io_cb)

    def __io_cb(self, chan, condition):
        if condition & gobject.IO_IN:
            line = chan.readline()
            if line:
                state = line.split(' ', 1)[0]
                self.emit('state-changed', state)
        return True
        

if __name__ == '__main__':
    def state_changed_cb(obj, state):
        print "*** ", state
    XScreenSaverMonitor().connect("state-changed", state_changed_cb)
    gobject.MainLoop().run()




import sys
import bonobo
from optparse import OptionParser

try:
    import dbus
except ImportError:
    HAVE_DBUS = False
else:
    HAVE_DBUS = True

def main():
    parser = OptionParser(usage="usage: %prog [options] [message]")
    parser.add_option("-f", "--full",
                      help="Message is in full format specification",
                      action="store_true", dest="full", default=False)
    parser.add_option("-s", "--stdin",
                      help="Read message from stdin;  Implies --full",
                      action="store_true", dest="stdin", default=False)
    if HAVE_DBUS:
        parser.add_option("", "--dbus",
                          help="Prefer D-BUS as IPC method",
                          action="store_true", dest="dbus", default=False)
    
    (options, args) = parser.parse_args()
    timeout = -1
    if len(args) == 2:
        print >> sys.stderr, "Warning: timeout argument is deprecated and will be removed in the future"
        timeout = int(args[1])
    if not options.stdin and not (1 <= len(args) <= 2):
        parser.error("wrong number of arguments")

    if options.stdin:
        msg = sys.stdin.read()
    else:
        msg = args[0]

    class ServerError(Exception):
        def __init__(self, value):
            self.value = value

    def send_dbus():
        osd = dbus.SessionBus().get_object("pt.inescporto.telecom.GnomeOSD", "/Server")
        if options.full or options.stdin:
            error = osd.showMessageFull(msg, dbus_interface="pt.inescporto.telecom.GnomeOSD")
        else:
            error = osd.showMessage(msg, timeout, dbus_interface="pt.inescporto.telecom.GnomeOSD")
        if error:
            raise ServerError(error)

    def send_bonobo():
        osd = bonobo.get_object("OAFIID:GNOME_OSD", "IDL:Bonobo/Application:1.0")
        osd.ref() # AppClient steals one reference
        osd = bonobo.AppClient(osd)

        if options.full or options.stdin:
            error = osd.msg_send("show-full", [msg])
        else:
            error = osd.msg_send("show", [msg, timeout])
        if error:
            raise ServerError(error)

    if HAVE_DBUS and options.dbus:
        order = [("D-BUS", send_dbus), ("Bonobo", send_bonobo)]
    else:
        order = [("Bonobo", send_bonobo)]
        if HAVE_DBUS:
            order.append(("D-BUS", send_dbus))

    for i, (name, func) in enumerate(order):
        if i > 0:
            print "Trying %s" % name
        try:
            func()
            break
        except ServerError, ex:
            print >> sys.stderr, "ServerError: %s" % ex.value
            raise SystemExit, 2
        except Exception, ex:
            print >> sys.stderr, "Error: %s" % ex
            continue



import gobject
import pango
import gtk
import gtk.gdk as gdk

BORDER_WIDTH=1

def osd(text, bgcolor, fgcolor, fontdesc,
        use_markup=False, alignment=pango.ALIGN_CENTER,
        fake_translucent_bg=False, drop_shadow=False,
        max_width=None, debug_frame=False, ellipsize=pango.ELLIPSIZE_NONE):
    assert isinstance(fontdesc, pango.FontDescription)
    win = gtk.Window(gtk.WINDOW_POPUP)
    win.add_events(gtk.gdk.ENTER_NOTIFY_MASK)
    darea = gtk.DrawingArea()
    win.add(darea)
    darea.show()
    if use_markup:
        layout = win.create_pango_layout('')
        layout.set_markup(text)
    else:
        layout = win.create_pango_layout(text)
    try:
        layout.set_ellipsize(ellipsize)
    except AttributeError:
        print "ellipsize attribute not supported, ignored"
    layout.set_justify(False)
    layout.set_alignment(alignment)
    layout.set_font_description(fontdesc)
    if max_width is None:
        MAX_WIDTH = gdk.screen_width() - 8
    else:
        MAX_WIDTH = max_width - 8
    layout.set_width(pango.SCALE*MAX_WIDTH)
    if ellipsize != pango.ELLIPSIZE_NONE:
        layout.set_wrap(pango.WRAP_WORD)
    width, height = layout.get_pixel_size()
    off_x = BORDER_WIDTH*2
    off_y = BORDER_WIDTH*2

    if alignment == pango.ALIGN_CENTER:
        off_x -= MAX_WIDTH/2 - width/2
    elif alignment == pango.ALIGN_RIGHT:
        off_x -= MAX_WIDTH - width
    
    width += BORDER_WIDTH*4
    height += BORDER_WIDTH*4
    if drop_shadow:
        drop_shadow_distance = max(2, int(fontdesc.get_size()/pango.SCALE*0.1))
        width += drop_shadow_distance
        height += drop_shadow_distance
    darea.set_size_request(width, height)
    darea.realize()
    pixmap = gtk.gdk.Pixmap(darea.window, width, height)
    #pixmap.set_colormap(darea.window.get_colormap())
    
    fg_gc = gdk.GC(pixmap); fg_gc.copy(darea.style.fg_gc[gtk.STATE_NORMAL])
    bg_gc = gdk.GC(pixmap); bg_gc.copy(darea.style.fg_gc[gtk.STATE_NORMAL])
    fg_gc.set_foreground(darea.get_colormap().alloc_color(fgcolor))
    bg_gc.set_background(darea.get_colormap().alloc_color(bgcolor))
    pixmap.draw_rectangle(bg_gc, True, 0, 0, width, height)
    pixmap.draw_layout(fg_gc, off_x, off_y, layout)

    if debug_frame:
        pixmap.draw_rectangle(fg_gc, False, 0, 0, width - 1, height - 1)

    bitmap = gtk.gdk.Pixmap(darea.window, width, height, 1)
    #bitmap.set_colormap(darea.window.get_colormap())
    
    fg_gc = gdk.GC(bitmap)
    bg_gc = gdk.GC(bitmap)
    fg_gc.set_foreground(gdk.Color(pixel=-1))
    bg_gc.set_background(gdk.Color(pixel=0))

    if fake_translucent_bg:
        w, h = 2, 2
        stipple = gtk.gdk.Pixmap(None, w, h, 1)
        stipple.draw_rectangle(bg_gc, True, 0, 0, w, h)
        stipple.draw_point(fg_gc, 0, 0)
        stipple.draw_point(fg_gc, 1, 1)
        fg_gc.set_stipple(stipple)
        bitmap.draw_rectangle(bg_gc, True, 0, 0, width, height)
        fg_gc.set_fill(gtk.gdk.STIPPLED)
        bitmap.draw_rectangle(fg_gc, True, 0, 0, width, height)
        fg_gc.set_fill(gtk.gdk.SOLID)

        # draw corners
        
        corner_size = int(fontdesc.get_size()/pango.SCALE*0.5)
        fg_gc.set_function(gtk.gdk.AND)
        gc = gdk.GC(bitmap)
        corner = gtk.gdk.Pixmap(bitmap, corner_size, corner_size)
        def draw_corner(angle1, x, y, arc_x0, arc_y0):
            gc.set_foreground(gdk.Color(pixel=0))
            corner.draw_rectangle(gc, True, 0, 0, corner_size, corner_size)
            gc.set_foreground(gdk.Color(pixel=1))
            corner.draw_arc(gc, True, arc_x0, arc_y0, corner_size*2, corner_size*2,
                            angle1*64, 90*64)
            bitmap.draw_drawable(fg_gc, corner, 0, 0, x, y, corner_size, corner_size)
        # top-left
        draw_corner(90, 0, 0, 0, 0)
        # bottom-left
        draw_corner(180, 0, height - corner_size, 0, -corner_size)
        # bottom-right
        draw_corner(270, width - corner_size, height - corner_size, -corner_size, -corner_size)
        # top-right
        draw_corner(0, width - corner_size, 0, -corner_size, 0)
        
        fg_gc.set_function(gtk.gdk.SET)
    else:
        bitmap.draw_rectangle(bg_gc, True, 0, 0, width, height)
    bitmap.draw_layout(fg_gc, off_x, off_y, layout)
    bitmap.draw_layout(fg_gc, off_x + BORDER_WIDTH, off_y, layout)
    bitmap.draw_layout(fg_gc, off_x + BORDER_WIDTH, off_y + BORDER_WIDTH, layout)
    bitmap.draw_layout(fg_gc, off_x, off_y + BORDER_WIDTH, layout)
    bitmap.draw_layout(fg_gc, off_x - BORDER_WIDTH, off_y + BORDER_WIDTH, layout)
    bitmap.draw_layout(fg_gc, off_x - BORDER_WIDTH, off_y, layout)
    bitmap.draw_layout(fg_gc, off_x - BORDER_WIDTH, off_y - BORDER_WIDTH, layout)
    bitmap.draw_layout(fg_gc, off_x, off_y - BORDER_WIDTH, layout)
    bitmap.draw_layout(fg_gc, off_x + BORDER_WIDTH, off_y - BORDER_WIDTH, layout)

    if drop_shadow:
        bitmap.draw_layout(fg_gc, off_x + drop_shadow_distance,
                           off_y + drop_shadow_distance, layout)
    if debug_frame:
        bitmap.draw_rectangle(fg_gc, False, 0, 0, width - 1, height - 1)

    darea.window.set_back_pixmap(pixmap, False)
    win.window.shape_combine_mask(bitmap, 0, 0)
    win.width = width
    win.height = height
    return win

def popup(text, bgcolor=None, fgcolor=None, fontdesc=None, use_markup=False):
    assert isinstance(fontdesc, pango.FontDescription)
    win = gtk.Window(gtk.WINDOW_POPUP)
    win.set_border_width(0)
    frame = gobject.new(gtk.Frame, shadow_type=gtk.SHADOW_ETCHED_OUT,
                        visible=True)
    win.add(frame)
    label = gobject.new(gtk.Label, use_markup=True, label=text, visible=True)
    label.modify_font(fontdesc)
    
    frame.add(label)

    win.show()
    win.width, win.height = win.allocation.width, win.allocation.height
    return win


if __name__ == '__main__':
    w = osd("<i>Hello</i> <span size='larger' weight='bold'>World</span>"
            "<span foreground='red'>!</span> sad adasd asd asd as das asd"
            " asdasdasdashdjasdlasl dkasdaskdjakaskdjasdlsajldj", "#000000", "#80ff80",
            pango.FontDescription("sans serif 48"),
            use_markup=True, drop_shadow=True, ellipsize=pango.ELLIPSIZE_END)
    w.move(gdk.screen_width()/2 - w.width/2, gdk.screen_height() - w.height - 10)
    def motion(wid, ev):
        print "enter notify!"
    w.connect("enter-notify-event", motion)
    w.show()
    gtk.main()




# -*- Mode: python; encoding: utf-8 -*-

## XChat OSD song notification
## (C) 2004 Gustavo J. A. M. Carneiro <gustavo@users.sf.net>
## Some code copied from xchatrb.py (xchat plugin), (c) Gustavo
## Carneiro and G-LiTe / <stephan@wilkogazu.nl>
## GPL license applies, blah blah

## More recent changes in main ChangeLog...
## Changes in version 0.3
## - Ditch gosd, require gnome-osd

## Changes in version 0.2
## - Thinner and less ugly border (Gustavo)

## Version: 0.1 (first release)

## Configuration variables:

SHOW_FULL_INFO = True

import pygtk; pygtk.require("2.0")
import ORBit
import CORBA
import bonobo.activation
import string
import bonobo
import gobject
import xml.sax.saxutils
from gnomeosd import gnome_osd_conf
import gettext
import gnome.ui
import sys
import os.path
import atexit
import gconfsync
from xscreensaver import XScreenSaverMonitor

try:
    import dbus
    import dbus.service
    if getattr(dbus, "version", (0,0,0)) >= (0,41,0):
        import dbus.glib
    HAVE_DBUS = True
except:
    HAVE_DBUS = False

# missing consts from the typelib
_ACTIVATION_FLAG_NO_LOCAL = 1<<0;      # No shared libraries
_ACTIVATION_FLAG_PRIVATE = 1<<1;       # start a new server and don't register it
_ACTIVATION_FLAG_EXISTING_ONLY = 1<<2; # don't start the server if not started


song_change_prefs = gconfsync.GConfSync("/apps/gnome-osd/plugins/song_change")
xchat_prefs = gconfsync.GConfSync("/apps/gnome-osd/plugins/xchat")
email_prefs = gconfsync.GConfSync("/apps/gnome-osd/plugins/email")
gaim_prefs = gconfsync.GConfSync("/apps/gnome-osd/plugins/gaim")
listener = None # Bonobo/ObjectDirectory:activation:register listener
screensaver_active = False

class OSD(object):
    def __init__(self):
        self.server = None
        self.client = None
        self.timeout = None
        self.last_message = None

    def _get_osd(self):
        if self.server is not None:
            try:
                # "ping" it
                self.server.ref()
                self.server.unref()
            except:
                print "OSD server died!"
                self.server = None
                self.client = None
        if self.server is None:
            self.server = bonobo.get_object("OAFIID:GNOME_OSD", "IDL:Bonobo/Application:1.0")
            self.server.ref() # AppClient steals one reference
            self.client = bonobo.AppClient(self.server)

    def _filter_timeout(self):
        self.timeout = None

    def send(self, message):
        if screensaver_active:
            return
        self._get_osd()
        if isinstance(message, unicode):
            message = message.encode("utf-8")
        if self.timeout is None:
            self.timeout = gobject.timeout_add(200, self._filter_timeout)
        else:
            if message == self.last_message:
                #print >> sys.stderr, "filtered duplicate message", message
                return
        err = self.client.msg_send("show-full", (message,))
        if err:
            print >> sys.stderr, err
        self.last_message = message

osd = OSD()

_rhythmbox = None
def get_rhythmbox():
    global _rhythmbox

    def get_new(): return bonobo.activation.activate(
	"repo_ids.has('IDL:GNOME/Rhythmbox:1.0')",
	[], _ACTIVATION_FLAG_EXISTING_ONLY)

    if _rhythmbox is None:
	_rhythmbox = get_new()
    else:
	# Check if the cached reference points to a server that is
	# still alive, otherwise ask bonobo-activation for a new fresh
	# reference.
	try:
	    # basically this means: _rhythmbox.ping()
	    _rhythmbox.ref()
	    _rhythmbox.unref()
	except CORBA.COMM_FAILURE:
	    _rhythmbox = get_new()
    return _rhythmbox

class SongInfo(object):
    __slots__ = ['title', 'album', 'artist', 'duration', 'track_number']


def get_psong():
    rb = get_rhythmbox()
    if rb is None:
	print "Rhythmbox is not running"
	return None
    pb = rb.getPlayerProperties()
    info = pb.getValue("song").value()
    pb.unref()
    if info is None:
        print "Rhythmbox is not playing anything"
        return None
    if not info.title:
        return "(untitled song)"
    return format_psong(info)

def format_psong(info):
    #  --- duration ---
    duration = info.duration
    if duration is None or duration == 0:
	duration_str = None
    else:
	if duration == -1:
	    print "Rhythmbox is not playing anything"
	    return None
	else:
	    hours    = duration // 3600
	    duration = duration % 3600
	    minutes  = duration // 60
	    duration = duration % 60
	    duration_str = str(minutes) + ":" + str(duration).zfill(2)
	    if hours == 0:
		duration_str = '%s:%s' % (str(minutes), str(duration).zfill(2))
	    else:
		duration_str = '%s:%s:%s' % (str(hours), str(minutes).zfill(2),
					     str(duration).zfill(2))
    # --- title ---
    title = escape(info.title)
    if title.endswith('.mp3') or title.endswith('.ogg'):
        title = title[:-4]
    msg = "♪ <span foreground='yellow'>%s</span>" % title
    if duration_str is not None:
        msg += " <span size='smaller'>(%s)</span> ♪" % duration_str
    else:
        msg += " ♪"

    extra_info = []
    # --- artist ---
    if info.artist:
        extra_info.append("<span size='smaller' style='italic'>%s</span>"
                          " <span foreground='yellow'>%s</span>" %
                          (_("Artist"), escape(info.artist)))

    # --- album ---
    if info.album:
        extra_info.append("<span size='smaller' style='italic'>%s</span>"
                          " <span foreground='yellow'>%s</span>" %
                          (_("Album"), escape(info.album)))
        if info.track_number != -1:
            extra_info.append("<span size='smaller' style='italic'>%s</span>"
                              " <span foreground='yellow'>%i</span>" %
                              (_("Track"), info.track_number))
    if SHOW_FULL_INFO and extra_info:
        return "\n".join((msg, "<span size='smaller'>%s</span>" % "   ".join(extra_info)))
    else:
        return msg


def rb_properties_cb(listener, event, foo, *ev):
    global song_change_prefs, osd
    if not song_change_prefs['enabled']:
        return
    song = get_psong()
    osd.send("<message id='rhythmbox' hide_timeout='10000'>" + song + "</message>")

def monitor_rhythmbox(*args):
    global listener
    rb = get_rhythmbox()
    if rb is not None:
        if listener is not None:
            bonobo.event_source_client_remove_listener(
                rb.getPlayerProperties(), listener)
        listener = bonobo.event_source_client_add_listener(
            rb.getPlayerProperties(), rb_properties_cb,
            "Bonobo/Property:change:song")

def die_cb(cli):
    print "Die!!"
    global listener
    rb = get_rhythmbox()
    if rb is not None:
        if listener is not None:
            bonobo.event_source_client_remove_listener(
                rb.getPlayerProperties(), listener)
    bonobo.main_quit()
    print "OK, quitting..."

def evolution_new_mail_callback(uri, folder=None):
    global email_prefs, osd
    if not email_prefs['enabled']:
        return
    if folder is None:
        if not uri.upper().endswith("INBOX"):
            return
    else:
        if folder.upper() != "INBOX":
            return
    osd.send("<message id='newmail' hide_timeout='3000'>You Have Mail</message>")

def muine_song_change_cb(arg):
    global song_change_prefs, osd
    if not song_change_prefs['enabled']:
        return
    d = dict((x.strip() for x in  s1.split(":", 1)) for s1 in arg.split("\n"))
    info = SongInfo()
    info.title = d['title']
    info.album = d['album']
    info.artist = d['artist']
    info.duration = int(d['duration'])
    info.track_number = int(d['track_number'])
    song = format_psong(info)
    osd.send("<message id='muine' hide_timeout='10000'>" + song + "</message>")

def escape(text):
    return xml.sax.saxutils.escape(text, {'\\': '&apos;',
                                          '"': '&quot;'})

XCHAT_EAT_NONE	= 0        # pass it on through!
XCHAT_EAT_XCHAT = 1        # don't let xchat see this event
XCHAT_EAT_PLUGIN = 2       # don't let other plugins see this event
XCHAT_EAT_ALL =	(XCHAT_EAT_XCHAT|XCHAT_EAT_PLUGIN) # don't let anything see this event

try:
    _xchat_ver = [int(x) for x in os.popen("xchat --version 2> /dev/null").read().split()[1].split('.')]
except IndexError:
    _xchat_ver = None

XCHAT_DBUS_SERVICE = "org.xchat.service"
if _xchat_ver >= [2,8]:
    XCHAT_DBUS_OBJECT = "/org/xchat/Remote"
    XCHAT_DBUS_INTERFACE = "org.xchat.plugin"
else:
    XCHAT_DBUS_OBJECT = "/org/xchat/RemoteObject"
    XCHAT_DBUS_INTERFACE = "org.xchat.interface"
del _xchat_ver

def monitor_xchat(bus):
    remote_object = bus.get_object(XCHAT_DBUS_SERVICE, XCHAT_DBUS_OBJECT)
    xchat = dbus.Interface(remote_object, XCHAT_DBUS_INTERFACE)
    event_ids = []
    for event in ('Channel Action Hilight', 'Channel Msg Hilight',
                  'Private Message', 'Private Message to Dialog'):
        id_ = xchat.HookPrint(event, 0, XCHAT_EAT_NONE)
        event_ids.append(id_)
        atexit.register(xchat.unhook, id_)
        
    def print_cb(word, id_, context_id=None):
        global xchat_prefs, osd
        if id_ not in event_ids:
            return
        if not xchat_prefs['enabled']:
            return
        sender = word[0]
        channel= xchat.GetInfo('channel')
        text = word[1]
        message = "<span foreground='red'>%s</span> "\
                  "<span weight='bold' foreground='white'>&lt;</span>%s"\
                  "<span weight='bold' foreground='white'>&gt;</span>" % \
                  (escape(channel), escape(sender))
        if xchat_prefs['full']:
            message += ' ' + escape(text)
        message = "<message id='xchat' ellipsize='end'>%s</message>" % message
        osd.send(message)
    
    remote_object.connect_to_signal("PrintSignal", print_cb,
                                    dbus_interface=XCHAT_DBUS_INTERFACE)


def xchat_registered_cb(*args):
    try:
        monitor_xchat(dbus.SessionBus())
    except dbus.DBusException, ex:
        print "XChat monitoring is not working:", ex

def screensaver_active_cb(active):
    global screensaver_active
    screensaver_active = active


def xscreensaver_state_changed(mon, state):
    global screensaver_active
    if state == 'UNBLANK':
        screensaver_active = False
    elif state == 'LOCK':
        screensaver_active = True

def monitor_xscreensaver():
    try:
        mon = XScreenSaverMonitor()
    except gobject.GError:
        print "xscreensaver not available"
    else:
        mon.connect("state-changed", xscreensaver_state_changed)


def rhythmbox_dbus_song_change_cb(uri):
    if not uri:
        return
    global song_change_prefs, osd
    if not song_change_prefs['enabled']:
        return
    bus = dbus.SessionBus()
    rbshellobj = bus.get_object('org.gnome.Rhythmbox', '/org/gnome/Rhythmbox/Shell')
    rbshell = dbus.Interface(rbshellobj, 'org.gnome.Rhythmbox.Shell')
    props = rbshell.getSongProperties(uri)
    info = SongInfo()
    info.title = props['title']
    info.album = props['album']
    info.artist = props['artist']
    info.duration = int(props['duration'])
    info.track_number = int(props['track-number'])
    song = format_psong(info)
    osd.send("<message id='rhythmbox' hide_timeout='10000'>" + song + "</message>")


def format_pgaim(buddy, action):
    return "<span foreground='yellow'>%s</span> " \
           "<span size='smaller' style='italic'>%s</span>" % (buddy, action)

def gaim_event_buddyon_cb(buddyid):
    global gaim_prefs, osd
    if not gaim_prefs['enabled']:
        return
    bus = dbus.SessionBus()
    gaimobj = bus.get_object("net.sf.gaim.GaimService", "/net/sf/gaim/GaimObject")
    gaim = dbus.Interface(gaimobj, "net.sf.gaim.GaimInterface")
    alias = gaim.GaimBuddyGetAlias(buddyid)
    message = format_pgaim(alias, "logs in")
    osd.send("<message id='gaim' hide_timeout='3000'>" + message + "</message>")

def gaim_event_buddyoff_cb(buddyid):
    global gaim_prefs, osd
    if not gaim_prefs['enabled']:
        return
    bus = dbus.SessionBus()
    gaimobj = bus.get_object("net.sf.gaim.GaimService", "/net/sf/gaim/GaimObject")
    gaim = dbus.Interface(gaimobj, "net.sf.gaim.GaimInterface")
    alias = gaim.GaimBuddyGetAlias(buddyid)
    message = format_pgaim(alias, "logs out")
    osd.send("<message id='gaim' hide_timeout='3000'>" + message + "</message>")

def gaim_event_message_cb(account, name):
    global gaim_prefs, osd
    if not gaim_prefs['enabled']:
        return
    bus = dbus.SessionBus()
    gaimobj = bus.get_object("net.sf.gaim.GaimService", "/net/sf/gaim/GaimObject")
    gaim = dbus.Interface(gaimobj, "net.sf.gaim.GaimInterface")
    buddy = gaim.GaimFindBuddy(account, name)
    if buddy != 0:
        alias = gaim.GaimBuddyGetAlias(buddy)
    else:
        alias = name
    message = format_pgaim(alias, "is writing a message")
    osd.send("<message id='gaim' hide_timeout='3000'>" + message + "</message>")

def gaim_event_messagereceived_cb(account, name, message, conversation, flags):
    global gaim_prefs, osd
    if not gaim_prefs['enabled']:
        return
    bus = dbus.SessionBus()
    gaimobj = bus.get_object("net.sf.gaim.GaimService", "/net/sf/gaim/GaimObject")
    gaim = dbus.Interface(gaimobj, "net.sf.gaim.GaimInterface")
    buddy = gaim.GaimFindBuddy(account, name)
    if buddy != 0:
        alias = gaim.GaimBuddyGetAlias(buddy)
    else:
        alias = name
    message = format_pgaim(alias, "sent a message")
    osd.send("<message id='gaim' hide_timeout='3000'>" + message + "</message>")

def monitor_gaim(bus):
    bus.add_signal_receiver(gaim_event_buddyon_cb, "BuddySignedOn",
                            "net.sf.gaim.GaimInterface")
    bus.add_signal_receiver(gaim_event_buddyoff_cb, "BuddySignedOff",
                            "net.sf.gaim.GaimInterface")
    bus.add_signal_receiver(gaim_event_message_cb, "BuddyTyping",
                            "net.sf.gaim.GaimInterface")
    bus.add_signal_receiver(gaim_event_messagereceived_cb, "ReceivedImMsg",
                            "net.sf.gaim.GaimInterface")


def monitor_dbus_apps():
    bus = dbus.SessionBus()
    bus.add_signal_receiver(evolution_new_mail_callback, "Newmail",
                            "org.gnome.evolution.mail.dbus.Signal")
    bus.add_signal_receiver(muine_song_change_cb, "SongChanged",
                            "org.gnome.Muine.Player")
    bus.add_signal_receiver(rhythmbox_dbus_song_change_cb, "playingUriChanged",
                            "org.gnome.Rhythmbox.Player")
    bus.add_signal_receiver(xchat_registered_cb, "NameOwnerChanged",
                            "org.freedesktop.DBus", arg0=XCHAT_DBUS_SERVICE)
    bus.add_signal_receiver(screensaver_active_cb, "ActiveChanged",
                            "org.gnome.ScreenSaver")
    monitor_gaim(bus)
    try:
        monitor_xchat(bus)
    except dbus.DBusException, ex:
        print "XChat monitoring is not working:", ex

if HAVE_DBUS:
    _our_bus_name = None

    def start_dbus_interface():
        global _our_bus_name
        session_bus = dbus.SessionBus()
        bus = dbus.SessionBus()
        if dbus.version >= (0,80):
            _our_bus_name = None
            # need to keep the BusName referenced in order to hang on to the name
            try:
                _our_bus_name = dbus.service.BusName("pt.inescporto.telecom.GnomeOSD.EventBridge",
                                                 bus, do_not_queue=True)
            except dbus.NameExistsException:
                raise SystemExit("already running")
        else:
            retval = dbus.dbus_bindings.bus_request_name(session_bus.get_connection(),
                                                         "pt.inescporto.telecom.GnomeOSD.EventBridge",
                                                         dbus.dbus_bindings.NAME_FLAG_DO_NOT_QUEUE)
            if retval in (dbus.dbus_bindings.REQUEST_NAME_REPLY_PRIMARY_OWNER,
                          dbus.dbus_bindings.REQUEST_NAME_REPLY_ALREADY_OWNER):
                pass
            elif retval in (dbus.dbus_bindings.REQUEST_NAME_REPLY_EXISTS,
                            dbus.dbus_bindings.REQUEST_NAME_REPLY_IN_QUEUE):
                raise SystemExit("already running")


def main():
    gettext.install("gnome-osd", gnome_osd_conf.datadir + "/locale")
    gnome.program_init(os.path.basename(sys.argv[0]), gnome_osd_conf.VERSION)
    cli = gnome.ui.master_client()
    cli.connect("die", die_cb)
    cli.set_restart_command(1, sys.argv[:1])
    global listener
    listener = bonobo.event_source_client_add_listener(
        bonobo.activation.activate("iid == 'OAFIID:Bonobo_Activation_EventSource'"),
        monitor_rhythmbox, "Bonobo/ObjectDirectory:activation:register")

    monitor_rhythmbox()
    monitor_xscreensaver()

    if HAVE_DBUS:
        monitor_dbus_apps()
        start_dbus_interface()
    else:
        print >> sys.stderr, "Unable to monitor D-BUS applications"

    bonobo.main()



# -*- python -*-
import pygtk
pygtk.require('2.0')
import bonobo
import xchat
import xml.sax.saxutils
import gconf
gconfclient = gconf.client_get_default()
gconfprefix = '/apps/xchatosd'

__module_name__ = 'xchatosd'
__module_version__ = '0.3'
__module_description__ = 'Notifies you of IRC message directed to you, using OSD'

def get_pref_bool(name, default):
    val = gconfclient.get('/'.join((gconfprefix, name)))
    if val is None:
        return default
    else:
        return val.get_bool()

def set_pref_bool(name, val):
    gconfclient.set_bool('/'.join((gconfprefix, name)), val)

ENABLED = get_pref_bool("enabled", True)
SHOW_FULL_MESSAGE = get_pref_bool("show_full_message", True)

def escape(text):
    return xml.sax.saxutils.escape(text, {'\\': '&apos;',
                                          '"': '&quot;'})

_osd = None
def get_osd():
    global _osd
    if _osd is not None:
        ser, cli = _osd
        try:
            # "ping" it
            ser.ref()
            ser.unref()
            return cli
        except:
            _osd = None
            print 'OSD server died!'
    if _osd is None:
        osd = bonobo.get_object('OAFIID:GNOME_OSD', 'IDL:Bonobo/Application:1.0')
        # AppClient steals one reference
        osd.ref()
        cli = bonobo.AppClient(osd)
        _osd = (osd, cli)
        return cli

def osd_show(text):
    osd = get_osd()
    if isinstance(text, unicode):
        text = text.encode('utf-8')
    err = osd.msg_send('show-full', (text,))
    if err:
        print err

def print_cb(word, word_eol, userdata):
    sender = word[0]
    channel= xchat.get_info('channel')
    text = word[1]
    message = "<span foreground='red'>%s</span> "\
              "<span weight='bold' foreground='white'>&lt;</span>%s"\
              "<span weight='bold' foreground='white'>&gt;</span>" % \
              (escape(channel), escape(sender))
    if SHOW_FULL_MESSAGE:
        message += ' ' + escape(text)
    message = "<message id='xchat' ellipsize='end'>%s</message>" % message
    if ENABLED:
        osd_show(message)
    return xchat.EAT_NONE

EVENTS = []
EVENTS.append(('Channel Action Hilight', 1))
EVENTS.append(('Channel Msg Hilight', 1))
EVENTS.append(('Private Message', 1))
EVENTS.append(('Private Message to Dialog', 1))
for event in EVENTS:
    xchat.hook_print(event[0], print_cb, event)

def unload(userdata):
    print 'Plugin ' + __module_name__ + ' ' + __module_version__ + ' unloaded.'

def osd_disable(word, word_eol, userdata):
    global ENABLED
    ENABLED = False
    set_pref_bool("enabled", ENABLED)
    print 'On Screen Display script disabled.'
    return xchat.EAT_ALL

def osd_enable(word, word_eol, userdata):
    global ENABLED
    ENABLED = True
    set_pref_bool("enabled", ENABLED)
    print 'On Screen Display script enabled.'
    return xchat.EAT_ALL

def osd_full(word, word_eol, userdata):
    global SHOW_FULL_MESSAGE
    SHOW_FULL_MESSAGE = True
    set_pref_bool("show_full_message", SHOW_FULL_MESSAGE)
    print 'On Screen Display script now displaying the message text.'
    return xchat.EAT_ALL

def osd_alert(word, word_eol, userdata):
    global SHOW_FULL_MESSAGE
    SHOW_FULL_MESSAGE = False
    set_pref_bool("show_full_message", SHOW_FULL_MESSAGE)
    print 'On Screen Display script now displaying only an OSD alert, without the message text.'
    return xchat.EAT_ALL

xchat.hook_unload(unload)
xchat.hook_command('OSD_ENABLE', osd_enable, help='/OSD_ENABLE Enable the on screen display script.')
xchat.hook_command('OSD_DISABLE', osd_disable, help='/OSD_DISABLE Disable the on screen display script.')
xchat.hook_command('OSD_FULLMESSAGE', osd_full, help='/OSD_FULLMESSAGE Display the full message text in the OSD.')
xchat.hook_command('OSD_ONLYALERT', osd_alert, help='/OSD_ONLYALERT Display only an OSD alert, without the message text.')

print 'Plugin ' + __module_name__ + ' ' + __module_version__ + ' loaded.'



import os.path
import gobject
import gtk
import gtk.gdk
import bonobo
import gconfsync
import gnome
import gnome.ui
import gettext
import locale
import gnome_osd_conf
import gtkexcepthook
try:
    import dbus
    if getattr(dbus, "version", (0,0,0)) >= (0,41,0):
        import dbus.glib
    HAVE_DBUS = True
except:
    HAVE_DBUS = False

class HIGContainer(gtk.Alignment):
    def __init__(self, title, child):
        gtk.Alignment.__init__(self)
        self.set(xalign=0, yalign=0, xscale=1, yscale=1)
        vbox = gtk.VBox()
        vbox.show()
        self.add(vbox)
        vbox.set_spacing(6)
        label = gtk.Label(); label.show()
        label.set_markup('<span weight="bold">%s</span>' % title)
        label.set_use_underline(True)
        label.set_alignment(0, 0.5)
        vbox.pack_start(label)
        hbox = gtk.HBox(); hbox.show()
        vbox.pack_start(hbox)
        label = gtk.Label('    '); label.show()
        hbox.pack_start(label, False, False)
        hbox.pack_start(child, True, True)
        self.show()
                 

class RadioGroup(gobject.GObject):
    __gsignals__ = {
        'changed': (gobject.SIGNAL_RUN_LAST,
                    None, # return type
                    (object, )) # parameter types
    }

    def __init__(self, items):
        self.__gobject_init__()
        self.radio_map = dict()
        self.items = dict()
        self.__ignore_toggled = False
        radio = None
        for item in items:
            radio = gtk.RadioButton(radio, item[0])
            self.radio_map[radio] = item[1]
            self.items[item[1]] = radio
            radio.show()
            radio.connect("toggled", self.__radio_toggled)

    def __radio_toggled(self, radio):
        if self.__ignore_toggled:
            return
        if radio.get_active():
            value = self.radio_map[radio]
            self.emit("changed", value)

    def get_value(self):
        for radio in self.radio_map:
            if radio.get_active():
                return self.radio_map[radio]

    def set_value(self, value):
        self.__ignore_toggled = True
        radio = self.items[value]
        radio.set_active(True)
        self.__ignore_toggled = False
if gobject.pygtk_version < (2,8):
    gobject.type_register(RadioGroup)

class RadioGroupSync(object):

    ''' Synchronizes a gconf key with a radio group '''

    def __init__(self, group, prefs, key):
        prefs.connect("changed::%s" % key, self.__pref_changed)
        group.connect("changed", self.__gui_changed)
        self.group = group
        self.prefs = prefs
        self.key = key
        group.set_value(prefs[key])

    def __pref_changed(self, prefs, key):
        self.group.set_value(prefs[key])

    def __gui_changed(self, unused_group, value):
        self.prefs[self.key] = value

class FontSync(object):
    def __init__(self, font, prefs, key):
        font.connect("font-set", self.__font_changed)
        prefs.connect("changed::%s" % key, self.__pref_changed)
        font.set_font_name(prefs[key])
        self.prefs = prefs
        self.key = key
        self.font = font
    def __font_changed(self, unused_font, font_name):
        self.prefs[self.key] = font_name
    def __pref_changed(self, prefs, key):
        self.font.set_font_name(prefs[key])

class ToggleSync(object):
    def __init__(self, toggle, prefs, key):
        self.prefs = prefs
        self.key = key
        self.toggle = toggle
        toggle.set_active(prefs[key])
        toggle.connect("toggled", self.__toggled)
        prefs.connect("changed::%s" % key, self.__pref_changed)
        self.__ignore_toggled = False
    def __toggled(self, toggle):
        if self.__ignore_toggled:
            return
        self.prefs[self.key] = toggle.get_active()
    def __pref_changed(self, prefs, key):
        self.__ignore_toggled = True
        self.toggle.set_active(prefs[key])
        self.__ignore_toggled = False

class AlignmentGroupSync(object):
    def __init__(self, group, prefs):
        self.group = group
        self.__pref_changed(prefs, None)
        prefs.connect("changed::osd_halignment", self.__pref_changed)
        prefs.connect("changed::osd_vposition", self.__pref_changed)
        self.prefs = prefs
        group.connect("changed", self.__gui_changed)

    def __pref_changed(self, prefs, unused_name):
        halign = prefs['osd_halignment']
        valign = prefs['osd_vposition']
        self.group.set_value("-".join((valign, halign)))
        
    def __gui_changed(self, unused_group, value):
        valign, halign = value.split('-')
        self.prefs['osd_halignment'] = halign
        self.prefs['osd_vposition'] = valign

def main():
    gettext.install("gnome-osd", os.path.join(gnome_osd_conf.datadir, "locale"))
    locale.setlocale(locale.LC_ALL, '')
    locale.bind_textdomain_codeset('gnome-osd','UTF-8')
    gettext.bind_textdomain_codeset('gnome-osd','UTF-8')
    gnome.program_init("gnome-osd-properties", gnome_osd_conf.VERSION)

    try:
        icon = gtk.gdk.pixbuf_new_from_file(os.path.join(gnome_osd_conf.datadir,
                                                         "pixmaps", "gnome-osd.png"))
        gtk.window_set_default_icon_list(icon)
        del icon
    except gobject.GError:
        print "Icon file not found"


    RESPONSE_TEST, RESPONSE_ABOUT = range(2)
    dialog = gtk.Dialog(title=_("OSD Properties"),
                        buttons=(_("_Test"), RESPONSE_TEST,
                                 gtk.STOCK_CLOSE, gtk.RESPONSE_CLOSE,
                                 )
                        )
    about_button = dialog.add_button(gtk.STOCK_ABOUT, RESPONSE_ABOUT)
    dialog.action_area.set_child_secondary(about_button, True)
    dialog.set_property("has-separator", 0)
    dialog.set_border_width(12)
    dialog.set_default_response(RESPONSE_TEST)
    dialog.set_resizable(False)

    def general_preferences():
        prefs = gconfsync.GConfSync("/apps/gnome-osd")
        main_vbox = gtk.VBox(False, 18)
        main_vbox.set_border_width(12)
        main_vbox.show()
        # Font
        hbox = gtk.HBox(False, 6)
        label = gtk.Label(_("OSD _Font: "))
        label.set_use_underline(True)
        label.set_alignment(0, 0.5)
        hbox.pack_start(label, False, True)

        font = gnome.ui.FontPicker()
        font.set_property('mode', gnome.ui.FONT_PICKER_MODE_FONT_INFO)
        hbox.pack_start(font, True, True)
        label.set_mnemonic_widget(font)
        hbox.show_all()
        main_vbox.add(hbox)

        FontSync(font, prefs, 'osd_font')

        hbox = gtk.HBox(False, 6)
        hbox.show()
        main_vbox.add(hbox)

        # Animations
        animations = gtk.CheckButton(_("_Animations"))
        animations.show()
        hbox.add(animations)
        ToggleSync(animations, prefs, "animations")

        # Drop shadow
        drop_shadow = gtk.CheckButton(_("_Drop shadow"))
        drop_shadow.show()
        hbox.add(drop_shadow)
        ToggleSync(drop_shadow, prefs, "drop_shadow")

        # Sound
        sound = gtk.CheckButton(_("_Sound"))
        sound.show()
        hbox.add(sound)
        ToggleSync(sound, prefs, "enable_sound")

        # Hide on hover
        hide = gtk.CheckButton(_("_Hide on mouse hover"))
        hide.show()
        main_vbox.add(hide)
        ToggleSync(hide, prefs, "hide_on_hover")

        # --- Alignment
        group = RadioGroup([
            (_("Top _left"), 'top-left'),
            (_("_Top"), 'top-center'),
            (_("Top _right"), 'top-right'),
            (_("L_eft"), 'center-left'),
            (_("_Center"), 'center-center'),
            (_("Ri_ght"), 'center-right'),
            (_("Bottom l_eft"), 'bottom-left'),
            (_("_Bottom"), 'bottom-center'),
            (_("Bottom r_ight"), 'bottom-right'),
            ])

        table = gtk.Table(3, 3)
        table.show()
        table.set_row_spacings(6)
        table.set_col_spacings(6)

        table.attach(group.items['top-left'], 0, 1, 0, 1)
        table.attach(group.items['top-center'], 1, 2, 0, 1)
        table.attach(group.items['top-right'], 2, 3, 0, 1)
        table.attach(group.items['center-left'], 0, 1, 1, 2)
        table.attach(group.items['center-center'], 1, 2, 1, 2)
        table.attach(group.items['center-right'], 2, 3, 1, 2)
        table.attach(group.items['bottom-left'], 0, 1, 2, 3)
        table.attach(group.items['bottom-center'], 1, 2, 2, 3)
        table.attach(group.items['bottom-right'], 2, 3, 2, 3)


        AlignmentGroupSync(group, prefs)


        # Avoid Panels
        avoid = gtk.CheckButton(_("A_void panels"))
        avoid.show()
        ToggleSync(avoid, prefs, "avoid_panels")

        vbox = gtk.VBox(False, 6)
        vbox.show()
        vbox.add(table)
        vbox.add(avoid)

        container = HIGContainer(_("Alignment"), vbox)
        main_vbox.add(container)
        return main_vbox


    def events_preferences():
        main_vbox = gtk.VBox(False, 18)
        main_vbox.set_border_width(12)
        main_vbox.show()
        tooltips = gtk.Tooltips()

        # Song change
        prefs = gconfsync.GConfSync("/apps/gnome-osd/plugins/song_change")
        check = gtk.CheckButton(_("Notify on _song changes"))
        tooltips.set_tip(check, _("Works with Rhythmbox and Muine"))
        check.show()
        container = HIGContainer(_("Music Player"), check)
        main_vbox.pack_start(container, False, False, 0)
        ToggleSync(check, prefs, "enabled")

        # XChat
        prefs = gconfsync.GConfSync("/apps/gnome-osd/plugins/xchat")
        vbox = gtk.VBox(False, 6)
        vbox.show()
        container = HIGContainer(_("IRC (Chat)"), vbox)
        main_vbox.pack_start(container, False, False, 0)
        
        check = gtk.CheckButton(_("Notify on _IRC messages directed to me"))
        tooltips.set_tip(check, _("Works with XChat (>= 2.6) with D-BUS plugin"))
        check.show()
        vbox.pack_start(check, False, False, 0)
        ToggleSync(check, prefs, "enabled")

        enabled_check = check
        check = gtk.CheckButton(_("Include _message text"))
        enabled_check.connect("toggled",
                              lambda check1, check2: check2.set_sensitive(check1.get_active()),
                              check)
        check.show()
        vbox.pack_start(check, False, False, 0)
        ToggleSync(check, prefs, "full")

        # Evolution
        prefs = gconfsync.GConfSync("/apps/gnome-osd/plugins/email")
        vbox = gtk.VBox(False, 6)
        vbox.show()
        container = HIGContainer(_("EMail"), vbox)
        main_vbox.pack_start(container, False, False, 0)
        
        check = gtk.CheckButton(_("Notify when new _email arrives"))
        tooltips.set_tip(check, _("Works with Evolution with enabled D-BUS plugin"))
        check.show()
        vbox.pack_start(check, False, False, 0)
        ToggleSync(check, prefs, "enabled")
        
        # gaim
        prefs = gconfsync.GConfSync("/apps/gnome-osd/plugins/gaim")
        vbox = gtk.VBox(False, 6)
        vbox.show()
        container = HIGContainer(_("Gaim"), vbox)
        main_vbox.pack_start(container, False, False, 0)
        
        check = gtk.CheckButton(_("Notify buddies actions"))
        tooltips.set_tip(check, _("Notify when a contact logs in or out, writes, or sends a message"))
        check.show()
        vbox.pack_start(check, False, False, 0)
        ToggleSync(check, prefs, "enabled")
        

        return main_vbox
    
    notebook = gtk.Notebook()
    notebook.show()
    dialog.vbox.add(notebook)

    notebook.append_page(general_preferences(), gobject.new(
        gtk.Label, label=_("General"), visible=True))
    notebook.append_page(events_preferences(), gobject.new(
        gtk.Label, label=_("Notifications"), visible=True))


    def dialog_response(dialog, response):
        if response == gtk.RESPONSE_CLOSE:
            dialog.destroy()
        elif response == RESPONSE_TEST:
            osd = bonobo.get_object("OAFIID:GNOME_OSD", "IDL:Bonobo/Application:1.0")
            osd.ref() # AppClient steals one reference
            osd = bonobo.AppClient(osd)
            osd.msg_send("show",
                         [_("This is a %s %s test sent from the preferences dialog")
                          % (gnome_osd_conf.PACKAGE, gnome_osd_conf.VERSION)])
        elif response == RESPONSE_ABOUT:
            about = gobject.new(gtk.AboutDialog, translator_credits=_("translator-credits"),
                                authors=["Gustavo J. A. M. Carneiro"],
                                version=gnome_osd_conf.VERSION, name="GNOME OSD",
                                license=(
"""GNOME OSD is free software; you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation; either version 2 of the License, or (at your option) any later version.

GNOME OSD is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along with GNOME OSD; if not, write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA"""),
                                wrap_license=True,
                                comments=_("An OSD (On Screen Display) message notification system for GNOME.")
                                )
            about.set_transient_for(dialog)
            about.show()

    dialog.connect("response", dialog_response)
    dialog.connect("destroy", lambda *unused_args: gtk.main_quit())

    if HAVE_DBUS:
        ## just in case, activate the event bridge now...
        dbus.SessionBus().start_service_by_name("pt.inescporto.telecom.GnomeOSD.EventBridge")

    dialog.show()
    gtk.main()

if __name__ == '__main__':
    main()



import pygtk; pygtk.require("2.0")
import sys
import gtk, pango
from gettext import gettext as _
from cStringIO import *
import traceback

_exception_in_progress = 0

def _info(type, value, tb):
    global _exception_in_progress
    if _exception_in_progress:
	# Exceptions have piled up, so we use the default exception
	# handler for such exceptions
	_excepthook_save(type, value, tb)
	return
    _exception_in_progress = 1
    dialog = gtk.MessageDialog(parent=None,
			       flags=0,
			       type=gtk.MESSAGE_WARNING,
			       buttons=gtk.BUTTONS_CLOSE,
			       message_format=_("A programming error has been detected"))
    dialog.format_secondary_text(_("It probably isn't fatal, but should be reported"
                                   " to the developers nonetheless."))
    dialog.set_title(_("Bug Detected"))
    dialog.set_default_response(gtk.RESPONSE_CLOSE)
    #dialog.set_border_width(12)
    #dialog.vbox.get_children()[0].set_spacing(12)

    # Details
    textview = gtk.TextView(); textview.show()
    textview.set_editable(False)
    textview.modify_font(pango.FontDescription("Monospace"))
    sw = gtk.ScrolledWindow(); sw.show()
    sw.set_policy(gtk.POLICY_AUTOMATIC, gtk.POLICY_AUTOMATIC)
    sw.add(textview)
    frame = gtk.Frame();
    frame.set_shadow_type(gtk.SHADOW_IN)
    frame.add(sw)
    frame.set_border_width(6)
    textbuffer = textview.get_buffer()
    trace = StringIO()
    traceback.print_exception(type, value, tb, None, trace)
    textbuffer.set_text(trace.getvalue())
    textview.set_size_request(gtk.gdk.screen_width()/2, gtk.gdk.screen_height()/3)
    frame.show()
    expander = gtk.Expander("Details")
    expander.add(frame)
    expander.show()
    dialog.vbox.add(expander)
    
    dialog.set_position(gtk.WIN_POS_CENTER)
    dialog.set_gravity(gtk.gdk.GRAVITY_CENTER)
    
    dialog.run()
    dialog.destroy()
    _exception_in_progress = 0
    
if not sys.stderr.isatty():
    _excepthook_save = sys.excepthook
    sys.excepthook = _info

if __name__ == '__main__':
    _excepthook_save = sys.excepthook
    sys.excepthook = _info
    print x + 1

    



"""
A pygtk based pseudo-thread (coroutines) framework

Introduction
============

  This module adds infrastructure for managing tasklets.  In this
  context, a X{tasklet} is defined as a routine that explicitly gives
  back control to the main program a certain points in the code, while
  waiting for certain events.  Other terms that may be used to describe
  tasklets include I{coroutines}, or I{cooperative threads}.

  The main advantages of tasklets are:

    - Eliminates the danger of unexpected race conditions or deadlocks
      that happen with preemptive (regular) threads;

    - Reduces the number of callbacks in your code, that sometimes are
      so many that you end up with I{spaghetti code}.

  The fundamental block used to create tasklets is Python's generators.
  Generators are objects that are defined as functions, and when called
  produce iterators that return values defined by the body of the
  function, specifically C{yield} statements.

  The neat thing about generators are not the iterators themselves but
  the fact that a function's state is completely frozen and restored
  between one call to the iterator's C{next()} and the following
  one. This allows the function to return control to a program's main
  loop while waiting for an event, such as IO on a socket, thus allowing
  other code to run in the mean time.  When the specified event occurs,
  the function regains control and continues executing as if nothing had
  happened.

Structure of a tasklet
======================

  At the outset, a tasklet is simply a python U{generator
  function<http://www.python.org/peps/pep-0255.html>}, i.e. a function
  or method containing one or more C{yield} statements.  Tasklets add a
  couple more requirements to regular generator functions:

    1. The values contained in C{yield} statements cannot be arbitrary
       (see below);

    2. After each C{yield} that indicates events, the function
       L{gtasklet.get_event} must be called to retrieve the event that
       just occurred.

Syntax for yield in tasklets
============================

  Inside tasklet functions, C{yield} statements are used to suspend
  execution of the tasklet while waiting for certain events.  Valid
  C{yield} values are:

    - A single L{Message} object, with a correctly set I{dest}
      parameter.  With this form, a message is sent to the indicated
      tasklet.  When C{yield} returns, no event is generated, so the
      tasklet should B{not} call L{get_event}.

    - One, or a sequence of:

       - A L{WaitCondition}, meaning to wait for that specific condition

       - A L{Tasklet}, with the same meaning as L{WaitForTasklet}C{(tasklet)}

       - A generator, with the same meaning as L{WaitForTasklet}C{(Tasklet(gen))}

      In this case, the tasklet is suspended until either one of the
      indicated events occurs.  The tasklet must call L{get_event} in
      this case.

Launching a tasklet
===================

  To start a tasklet, the L{Tasklet} constructor must be used::
    import gtasklet

    def my_task(x):
        [...]

    gtasklet.Tasklet(my_task(x=0))

  Alternatively, L{gtasklet.run} can be used to the same effect::
    gtasklet.run(my_task(x=0))

Examples
========

  Background timeout task
  -----------------------
    This example demonstrates basic tasklet structure and timeout events::
      import gobject
      import gtasklet

      mainloop = gobject.MainLoop()

      def simple_counter(numbers):
          timeout = gtasklet.WaitForTimeout(1000)
          for x in xrange(numbers):
              print x
              yield timeout
              gtasklet.get_event()
          mainloop.quit()

      gtasklet.run(simple_counter(10))
      mainloop.run()

  Message passing
  ---------------
    This example extends the previous one and demonstrates message passing::

      import gobject
      import gtasklet

      mainloop = gobject.MainLoop()

      def printer():
          msgwait = gtasklet.WaitForMessages(accept=("quit", "print"))
          while True:
              yield msgwait
              msg = gtasklet.get_event()
              if msg.name == "quit":
                  return
              assert msg.name == 'print'
              print ">>> ", msg.value

      def simple_counter(numbers, task):
          timeout = gtasklet.WaitForTimeout(1000)
          for x in xrange(numbers):
              yield gtasklet.Message('print', dest=task, value=x)
              yield timeout
              gtasklet.get_event()
          yield gtasklet.Message('quit', dest=task)
          mainloop.quit()

      task = gtasklet.run(printer())
      gtasklet.run(simple_counter(10, task))
      mainloop.run()


@author: Gustavo J. A. M. Carneiro
@organization: INESC Porto
@copyright: Gustavo J. A. M. Carneiro
@license: GNU LGPL
@contact: U{mailto:gjc@inescporto.pt}

"""

__revision__ = (0, 4, 0)

import gobject
import warnings
import types

assert gobject.pygtk_version >= (2, 8)

_event = None

def get_event():
    """
    Return the last event that caused the current tasklet to regain control.

    @warning: this function should be called exactly once after each
    yield that includes a wait condition.

    """
    global _event
    assert _event is not None
    event = _event
    _event = None
    return event

def run(gen):
    """Start running a generator as a L{Tasklet}.

    @parameter gen: generator object that implements the tasklet body.
    @return: a new L{Tasklet} instance, already running.

    @note: this is strictly equivalent to calling C{Tasklet(gen)}.

    """
    return Tasklet(gen)


class WaitCondition(object):
    '''
    Base class for all wait-able condition objects.

    WaitConditions are used in a yield statement inside tasklets body
    for specifying what event(s) it should wait for in order to
    receive control once more.'''
    
    def __init__(self):
        '''Abstract base class, do not call directly'''
        self.triggered = False

    def arm(self, tasklet):
        '''Prepare the wait condition to receive events.

        When a wait condition receives the event it is waiting for, it
        should call the method
        L{wait_condition_fired<Tasklet.wait_condition_fired>} of the
        tasklet with the wait condition as argument.  The method
        returns True or False; if it returns True, it means the
        WaitCondition object must "rearm" itself (continue to monitor
        events), otherwise it should disarm.

        @parameter tasklet: the tasklet instance the wait condition is
        to be associated with.

        @attention: this method normally should not be called directly
        by the programmer.

        '''
        raise NotImplementedError

    def disarm(self):
        '''Stop the wait condition from receiving events.

        @attention: this method normally should not be called by the
        programmer.'''
        raise NotImplementedError


class WaitForIO(WaitCondition):
    '''An object that waits for IO conditions on sockets or file
    descriptors.

    

    '''
    def __init__(self, filedes, condition=gobject.IO_IN,
                 priority=gobject.PRIORITY_DEFAULT):
        '''
          @param filedes: object to monitor for IO
          @type filedes: int file descriptor, or a
          L{gobject.IOChannel}, or an object with a C{fileno()}
          method, such as socket or unix file.

          @param condition: IO event mask
          @type condition: a set of C{gobject.IO_*} flags ORed together
          @param priority: mainloop source priority

        '''
        
        WaitCondition.__init__(self)
        self.filedes = filedes
        self.__condition = condition # listen condition
        self.condition = None # last occurred condition
        self.__callback = None
        self.__id = None
        self.__priority = priority

    def arm(self, tasklet):
        '''Overrides WaitCondition.arm'''
        self.__callback = tasklet.wait_condition_fired
        if self.__id is None:
            try:
                ## http://bugzilla.gnome.org/show_bug.cgi?id=139176
                iochan = isinstance(self.filedes, gobject.IOChannel)
            except AttributeError:
                iochan = False
            if iochan:
                self.__id = self.filedes.add_watch(self.__condition,
                                                   self.__io_cb,
                                                   priority=self.__priority)
            else:
                if isinstance(self.filedes, int):
                    filedes = self.filedes
                else:
                    filedes = self.filedes.fileno()
                self.__id = gobject.io_add_watch(filedes, self.__condition,
                                                 self.__io_cb,
                                                 priority=self.__priority)

    def disarm(self):
        '''Overrides WaitCondition.disarm'''
        if self.__id is not None:
            gobject.source_remove(self.__id)
            self.__id = None
            self.__callback = None

    def __io_cb(self, unused_filedes, condition):
        self.triggered = True
        self.condition = condition
        retval = self.__callback(self)
        self.triggered = False
        if not retval:
            self.__id = None
        return retval


class WaitForTimeout(WaitCondition):
    '''An object that waits for a specified ammount of time (a timeout)'''
    def __init__(self, timeout, priority=gobject.PRIORITY_DEFAULT):
        '''An object that waits for a specified ammount of time.

        @param timeout: ammount of time to wait, in miliseconds
        @param priority: mainloop priority for the timeout event

        '''
        
        WaitCondition.__init__(self)
        self.timeout = timeout
        self.__id = None
        self.__tasklet = None
        self.__priority = priority
        
    def arm(self, tasklet):
        '''See WaitCondition.arm'''
        if self.__id is None:
            self.__tasklet = tasklet
            self.__id = gobject.timeout_add(self.timeout, self.__timeout_cb,
                                            priority=self.__priority)

    def disarm(self):
        '''See WaitCondition.disarm'''
        if self.__id is not None:
            gobject.source_remove(self.__id)
            self.__id = None
            self.__tasklet = None

    def restart(self):
        '''Restart the timeout.  Makes time counting start again from zero.'''
        tasklet = self.__tasklet
        self.disarm()
        self.arm(tasklet)

    def __timeout_cb(self):
        assert self.__id is not None,\
               "timeout called even after being unregistered"
        self.triggered = True
        retval = self.__tasklet.wait_condition_fired(self)
        assert retval is not None
        self.triggered = False
        if not retval:
            self.__id = None
        return retval

class WaitForIdle(WaitCondition):
    '''An object that waits for the main loop to become idle'''

    def __init__(self, priority=gobject.PRIORITY_DEFAULT_IDLE):
        '''An object that waits for the main loop to become idle, with a
        priority indicated by @priority'''
        WaitCondition.__init__(self)
        self.__callback = None
        self.__id = None
        self.__priority = priority

    def arm(self, tasklet):
        '''See WaitCondition.arm'''
        if self.__id is None:
            self.__callback = tasklet.wait_condition_fired
            self.__id = gobject.idle_add(self.__idle_cb, self.__priority)

    def disarm(self):
        '''See WaitCondition.disarm'''
        if self.__id is not None:
            gobject.source_remove(self.__id)
            self.__id = None
            self.__callback = None

    def __idle_cb(self):
        self.triggered = True
        retval = self.__callback(self)
        self.triggered = False
        if not retval:
            self.__id = None
        return retval


class WaitForTasklet(WaitCondition):
    '''An object that waits for a tasklet to complete'''
    def __init__(self, tasklet):
        '''An object that waits for another tasklet to complete'''
        
        WaitCondition.__init__(self)
        self.__tasklet = tasklet
        self.__id = None
        self.__idle_id = None
        self.__callback = None
        self.retval = None
        
    def arm(self, tasklet):
        '''See WaitCondition.arm'''
        self.__callback = tasklet.wait_condition_fired
        if self.__id is None:
            self.__id = self.__tasklet.add_join_callback(self.__join_cb)

    def disarm(self):
        '''See WaitCondition.disarm'''
        if self.__idle_id is not None:
            gobject.source_remove(self.__idle_id)
            self.__idle_id = None
        if self.__id is not None:
            self.__tasklet.remove_join_callback(self.__id)
            self.__id = None
            self.__callback = None

    def __join_cb(self, tasklet, retval):
        assert tasklet is self.__tasklet
        assert self.__idle_id is None
        self.__id = None
        self.__idle_id = gobject.idle_add(self.__idle_cb)
        self.retval = retval

    def __idle_cb(self):
        self.triggered = True
        self.__callback(self)
        self.triggered = False
        self.__tasklet = None
        self.__callback = None
        self.__id = None
        self.__idle_id = None
        return False

class WaitForSignal(WaitCondition):
    '''An object that waits for a signal emission'''

    def __init__(self, obj, signal):
        '''Waits for a signal to be emitted on a specific GObject instance.

        @param obj: object monitor for the signal
        @type obj: gobject.GObject
        @param signal: signal name
        @type signal: str

        '''
        WaitCondition.__init__(self)
        assert isinstance(obj, gobject.GObject)
        assert isinstance(signal, str)
        self.object = obj
        self.signal = signal
        self.__callback = None
        self.__id = None
        self.__destroy_id = None
        self.signal_args = None

    def arm(self, tasklet):
        '''See WaitCondition.arm'''
        if self.__id is None:
            self.__callback = tasklet.wait_condition_fired
            self.__id = self.object.connect(self.signal, self.__signal_cb)
            if gobject.signal_lookup("destroy", self.object):
                self.__destroy_id = self.object.connect("destroy",
                                                        self.__object_destroyed)

    def __object_destroyed(self, obj):
        self.object = None
        self.__id = None
        self.__destroy_id = None
        self.__callback = None

    def disarm(self):
        '''See WaitCondition.disarm'''
        if self.__id is not None:
            self.object.disconnect(self.__id)
            self.__id = None
            self.__callback = None
        if self.__destroy_id is not None:
            self.object.disconnect(self.__destroy_id)
            self.__destroy_id = None

    def __signal_cb(self, obj, *args):
        assert obj is self.object
        self.triggered = True
        self.signal_args = args
        retval = self.__callback(self)
        self.triggered = False
        if not retval:
            self.__id = None
        return retval


class WaitForProcess(WaitCondition):
    '''An object that waits for a process to end'''
    def __init__(self, pid):
        '''
        Creates an object that waits for a subprocess

        @parameter pid: Process identifier
        @type pid: int
        '''
        WaitCondition.__init__(self)
        self.pid = pid
        self.__callback = None
        self.__id = None
        self.status = None

    def arm(self, tasklet):
        '''See WaitCondition.arm'''
        self.__callback = tasklet.wait_condition_fired
        if self.__id is None:
            self.__id = gobject.child_watch_add(self.pid, self.__child_cb)

    def disarm(self):
        '''See WaitCondition.disarm'''
        if self.__id is not None:
            gobject.source_remove(self.__id)
            self.__id = None
            self.__callback = None

    def __child_cb(self, unused_pid, status):
        self.triggered = True
        self.status = status
        self.__callback(self)
        self.triggered = False
        self.status = None
        self.__id = None


class Message(object):
    '''A message that can be received by or sent to a tasklet.'''
    
    __slots__ = 'name', 'dest', 'value', 'sender'

    ACCEPT, DEFER, DISCARD = range(3)
    
    def __init__(self, name, dest=None, value=None, sender=None):
        '''
          @param name: name of message
          @type name: str
          @param dest: destination tasklet for this message
          @type dest: L{Tasklet}
          @param value: value associated with the message
          @param sender: sender tasklet for this message
          @type sender: L{Tasklet}

        '''
        assert isinstance(sender, (Tasklet, type(None)))
        assert isinstance(dest, (Tasklet, type(None)))
        assert isinstance(name, basestring)
        self.name = name
        self.value = value
        self.sender = sender
        self.dest = dest
        
#     def get_name(self):
#         """Return the message name"""
#         return self.name
#     def get_value(self):
#         """Return the message value"""
#         return self.value
#     def get_sender(self):
#         """Return the message sender"""
#         return self.sender
#     def get_dest(self):
#         """Return the message destination"""
#         return self.dest


def _normalize_list_argument(arg, name):
    """returns a list of strings from an argument that can be either
    list of strings, None (returns []), or a single string returns
    ([arg])"""
    
    if arg is None:
        return []
    elif isinstance(arg, basestring):
        return [arg]
    elif isinstance(arg, (list, tuple)):
        return arg
    raise TypeError("Argument '%s' must be None, a string, or "
                    "a sequence of strings, not %r" % (name, type(arg)))


class WaitForMessages(WaitCondition):
    '''An object that waits for messages to arrive'''
    def __init__(self, accept=None, defer=None, discard=None):
        '''Creates an object that waits for a set of messages to
        arrive.

        @warning: unlike other wait conditions, when a message
        is received, a L{Message} instance is returned by L{get_event()},
        not the L{WaitForMessages} instance.

        @param accept: message name or names to accept (receive) in
        the current state
        @type accept: string or sequence of string

        @param defer: message name or names to defer (queue) in the
        current state
        @type defer: string or sequence of string

        @param discard: message name or names to discard (drop) in the
        current state
        @type discard: string or sequence of string

        '''
        WaitCondition.__init__(self)
        self.__tasklet = None
        accept = _normalize_list_argument(accept, 'accept')
        defer = _normalize_list_argument(defer, 'defer')
        discard = _normalize_list_argument(discard, 'discard')
        self.actions = dict()
        for name in accept:
            self.actions[name] = Message.ACCEPT
        for name in defer:
            self.actions[name] = Message.DEFER
        for name in discard:
            self.actions[name] = Message.DISCARD
        
    def arm(self, tasklet):
        '''Overrides WaitCondition.arm'''
        self.__tasklet = tasklet
        tasklet.message_actions.update(self.actions)

    def disarm(self):
        '''Overrides WaitCondition.disarm'''
        assert self.__tasklet is not None
        for name in self.actions:
            del self.__tasklet.message_actions[name]


class Tasklet(object):
    '''An object that launches and manages a tasklet.'''

    STATE_RUNNING, STATE_SUSPENDED, STATE_MSGSEND = range(3)
    
    def __init__(self, gen=None):
        '''
        Launch a generator tasklet.

        @param gen: a generator object that implements the tasklet main body

        If `gen` is omitted or None, L{run} should be overridden in a
        subclass.

        '''
        self.__event = None
        self.__join_callbacks = {}
        self.wait_list = []
        self.__message_queue = []
        self._message_actions = {}
        self.state = Tasklet.STATE_SUSPENDED
        if gen is None:
            self.gen = self.run()
        else:
            assert isinstance(gen, types.GeneratorType)
            self.gen = gen
        self.__next_round() # bootstrap

    def get_message_actions(self):
        """Dictionary mapping message names to actions ('accept' or
           'discard' or 'defer').  Should normally not be accessed
           directly by the programmer.

        """
        return self._message_actions

    message_actions = property(get_message_actions)
        
    def run(self):
        """
        Method that executes the task.

        Should be overridden in a subclass if no generator is passed
        into the constructor."""
        raise ValueError("Should be overridden in a subclass "
                         "if no generator is passed into the constructor")

    def __invoke(self):
        global _event
        assert _event is None
        had_event = (self.__event is not None)
        _event = self.__event
        try:
            self.state = Tasklet.STATE_RUNNING
            gen_value = self.gen.next()
            self.state = Tasklet.STATE_SUSPENDED
            assert gen_value is not None
        except StopIteration, ex:
            if ex.args:
                retval, = ex.args
            else:
                retval = None
            self.__join(retval)
            return None
        if __debug__:
            if had_event and _event is not None:
                warnings.warn("Tasklet %s forgot to read an event!" % self)
        self.__event = None
        return gen_value

    def __next_round(self):
        assert self.state == Tasklet.STATE_SUSPENDED
        old_wait_list = self.wait_list
        while True: # loop while tasklet yields tasklet.post_message(...)

            gen_value = self.__invoke()
            if gen_value is None:
                return

            if isinstance(gen_value, Message):
                msg = gen_value
                self.state = Tasklet.STATE_MSGSEND
                msg.dest.send_message(msg)
                continue # loop because we posted a message
            elif isinstance(gen_value, tuple):
                self.wait_list = list(gen_value)
            elif isinstance(gen_value, list):
                self.wait_list = gen_value
            else:
                self.wait_list = [gen_value]
            
            for i, val in enumerate(self.wait_list):
                if isinstance(val, WaitCondition):
                    continue
                elif isinstance(val, types.GeneratorType):
                    self.wait_list[i] = WaitForTasklet(Tasklet(val))
                elif isinstance(val, Tasklet):
                    self.wait_list[i] = WaitForTasklet(val)
                else:
                    raise TypeError("yielded values must be WaitConditions,"
                                    " generators, or a single Message")

            self._update_wait_conditions(old_wait_list)

            msg = self._dispatch_message()
            if msg is not None:
                self.__event = msg
                continue ## send a message

            break

    def _dispatch_message(self):
        '''get next message that a tasklet wants to receive; discard
        messages that should be discarded'''
        ## while sending out messages, the tasklet implicitly queues
        ## all incoming messages
        if self.state == Tasklet.STATE_MSGSEND:
            return None

        ## filter out messages with discard action
        def __get_action(msg):
            try:
                return self._message_actions[msg.name]
            except KeyError:
                warnings.warn("Implicitly discarding message %s"
                              " directed to tasklet %s" % (msg, self))
                return Message.DISCARD
        if __debug__:
            self.__message_queue = [msg for msg in self.__message_queue
                                    if __get_action(msg) != Message.DISCARD]
        else:
            ## slightly more efficient version of the above
            self.__message_queue = [msg for msg in self.__message_queue
                if (self._message_actions.getdefault(msg.name, Message.DISCARD)
                    != Message.DISCARD)]

        ## find next ACCEPT-able message from queue, and pop it out
        for idx, msg in enumerate(self.__message_queue):
            if self._message_actions[msg.name] == Message.ACCEPT:
                break
        else:
            return None
        return self.__message_queue.pop(idx)
    

    def _update_wait_conditions(self, old_wait_list):
        '''disarm wait conditions removed and arm new wait conditions'''
        
        ## disarm conditions removed from the wait list
        for cond in old_wait_list:
            if cond not in self.wait_list:
                cond.disarm()
        
        ## arm the conditions added to the wait list
        for cond in self.wait_list:
            if cond not in old_wait_list:
                cond.arm(self)

    def wait_condition_fired(self, triggered_cond):
        """Method that should be called when a wait condition fires"""
        assert triggered_cond in self.wait_list
        assert self.__event is None
        self.__event = triggered_cond
        self.__next_round()
        self.__event = None
        if self.wait_list is None:
            return False
        else:
            return (triggered_cond in self.wait_list)

    def add_join_callback(self, callback):
        '''
        Add a callable to be invoked when the tasklet finishes.
        Return a connection handle that can be used in
        remove_join_callback()

        The callback will be called like this::
              callback(tasklet, retval)
        where tasklet is the tasklet that finished, and retval its
        return value (or None).

        When a join callback is invoked, it is automatically removed,
        so calling L{remove_join_callback} afterwards produces a KeyError
        exception.

        '''
        handle = hash(callback)
        while handle in self.__join_callbacks: # handle collisions
            handle += 1
        self.__join_callbacks[handle] = callback
        return handle

    def remove_join_callback(self, handle):
        '''Remove a join callback previously added with L{add_join_callback}'''
        del self.__join_callbacks[handle]

    def __join(self, retval):
        for cond in self.wait_list:
            cond.disarm()
        self.gen = None
        self.wait_list = []
        callbacks = self.__join_callbacks.values()
        self.__join_callbacks.clear()
        for callback in callbacks:
            callback(self, retval)


    def send_message(self, message):
        """Send a message to be received by the tasklet as an event.

        @warning: Don't call this from another tasklet, only from the
        main loop!  To send a message from another tasklet, yield a
        L{Message} with a correctly set 'dest' parameter.

        """
        assert isinstance(message, Message)
        assert self.__event is None
        if message.dest is None:
            message.dest = self
        self.__message_queue.append(message)
        self.__event = self._dispatch_message()
        if self.__event is not None:
            self.__next_round()


## ----------------------------
## And here's an example...
## ----------------------------
class _CountSomeNumbers2(Tasklet):
    '''Counts numbers with at random time spacings'''

    def __init__(self, count, timeout):
        """foo"""
        self.count = count
        self.timeout = timeout
        Tasklet.__init__(self)
    
    def run(self):
        '''Execute the task.'''
        for i in xrange(self.count):
            print ">> _count_some_numbers2", i
            yield (WaitForTimeout(random.randint(70, self.timeout)),
                   WaitForMessages(accept='quit'))
            event = get_event()
            if isinstance(event, Message) and event.name == 'quit':
                ## this would be the place to do some cleanup.
                return
        raise StopIteration(self.count*2)

def _count_some_numbers1(count):
    '''Counts numbers with at fixed time spacings'''
    timeout = WaitForTimeout(1000)
    for i in xrange(count):
        print "_count_some_numbers1", i
        task2 = _CountSomeNumbers2(10, 130)
        yield timeout, task2
        event = get_event()
        if event is timeout:
            print ">>> Got tired of waiting for task!! Canceling!"
            ## send a message asking the tasklet to stop
            yield Message('quit', dest=task2)
        elif isinstance(event, WaitForTasklet):
            print ">>> task returned %r, good task!" % event.retval
            ## restart timeout from scratch, otherwise it keeps
            ## running and we end up giving the next task too little
            ## time.
            timeout.restart()
        else:
            assert False, "strange event"


def _test():
    '''a simple test/example'''
    Tasklet(_count_some_numbers1(100))
    gobject.MainLoop().run()


if __name__ == '__main__':
    import random
    _test()



import gobject
import gconf

## Please take note of this important restriction: all keys in the
## gconf dir being monitored must have a corresponding schema default,
## or else this might break


class GConfSync(gobject.GObject):
    __gsignals__ = {
        'changed': (gobject.SIGNAL_RUN_LAST|gobject.SIGNAL_DETAILED,
                    None, # return type
                    (str, )) # parameter types
    }

    def __init__(self, prefix, enums={}):
        gobject.GObject.__init__(self)
        self.prefix = prefix
        gc = gconf.client_get_default()
        keys = dict() # mapping key -> type
        for entry in gc.all_entries(prefix):
            assert entry.key[:len(prefix)] == prefix
            keyname = entry.key[len(prefix):].lstrip('/')
            keys[keyname] = entry.value.type
        self.valid_keys = keys
        self.__key_just_been_set = None
        self.enums_mapping = enums
        reverse_map = dict()
        for key in enums:
            reverse_map[key] = dict(zip(enums[key].values(), enums[key].keys()))
        self.enums_reverse_mapping = reverse_map
        gc.add_dir(prefix, gconf.CLIENT_PRELOAD_ONELEVEL)
        gc.notify_add(prefix, self.__gconfclientnotifyfunc)

    def __gconfclientnotifyfunc(self, unused_gc, unused_cnxn_id, entry, unused_user_data):
        assert entry.key[:len(self.prefix)] == self.prefix
        keyname = entry.key[len(self.prefix):].lstrip('/')
        if keyname and keyname != self.__key_just_been_set:
            self.emit("changed::" + keyname, keyname)
        self.__key_just_been_set = None
        
    def __value_to_python(value):
        if value.type == gconf.VALUE_BOOL:
            if value.get_bool():
                return True
            else:
                return False
        elif value.type == gconf.VALUE_FLOAT:
            return value.get_float()
        elif value.type == gconf.VALUE_INT:
            return value.get_int()
        elif value.type == gconf.VALUE_INVALID:
            raise ValueError, "GConfValue of invalid type"
        elif value.type == gconf.VALUE_LIST:
            return map(GConfSync.__value_to_python, value.get_list())
        elif value.type == gconf.VALUE_PAIR:
            raise NotImplementedError
        elif value.type == gconf.VALUE_SCHEMA:
            raise NotImplementedError
        elif value.type == gconf.VALUE_STRING:
            return value.get_string()
        else:
            return None
    
    __value_to_python = staticmethod(__value_to_python)

    def __gconf_get(self, key):
        prefix = self.prefix
        val = gconf.client_get_default().get("/".join((prefix, key)))
        val = self.__value_to_python(val)
        if key in self.enums_mapping:
            val = self.enums_mapping[key][val]
        return val

    def __gconf_set(self, key, val):
        prefix = self.prefix
        gc = gconf.client_get_default()
        value_type = self.valid_keys[key]
        self.__key_just_been_set = key
        if key in self.enums_reverse_mapping:
            val = self.enums_reverse_mapping[key][val]
        
        key = '/'.join((prefix, key))

        if value_type == gconf.VALUE_BOOL:
            gc.set_bool(key, val)
        elif value_type == gconf.VALUE_FLOAT:
            gc.set_float(key, val)
        elif value_type == gconf.VALUE_INT:
            gc.set_int(key, val)
        elif value_type == gconf.VALUE_INVALID:
            raise ValueError, "GConfValue of invalid type"
        elif value_type == gconf.VALUE_LIST:
            raise NotImplementedError
        elif value_type == gconf.VALUE_PAIR:
            raise NotImplementedError
        elif value_type == gconf.VALUE_SCHEMA:
            raise NotImplementedError
        elif value_type == gconf.VALUE_STRING:
            gc.set_string(key, val)

    def __getitem__(self, key):
        return self.__gconf_get(key)

    def __setitem__(self, key, val):
        self.__gconf_set(key, val)

    def parse_string(self, key, val):
        value_type = self.valid_keys[key]

        if value_type == gconf.VALUE_BOOL:
            if val.lower() in ('true', 'enable', 'on'):
                return True
            elif val.lower() in ('false', 'disable', 'off'):
                return False
            else:
                raise ValueError, "unable to parse '%s' as boolean" % val 
        elif value_type == gconf.VALUE_STRING:
            if key in self.enums_mapping:
                return self.enums_mapping[key][val]
            else:
                return val
        elif value_type == gconf.VALUE_FLOAT:
            return float(val)
        elif value_type == gconf.VALUE_INT:
            return int(val)
        elif value_type == gconf.VALUE_INVALID:
            raise ValueError, "GConfValue of invalid type"
        elif value_type == gconf.VALUE_LIST:
            raise NotImplementedError, "parsing list gconf values not implemented"
        elif value_type == gconf.VALUE_PAIR:
            raise NotImplementedError, "parsing pair gconf values not implemented"
        elif value_type == gconf.VALUE_SCHEMA:
            raise NotImplementedError, "parsing schema gconf values not implemented"
        raise NotImplementedError, "parsing <unknown> gconf values not implemented"

if gobject.pygtk_version < (2,8):
    gobject.type_register(GConfSync)


if __name__ == '__main__':
    import gtk, pango
    prefs = GConfSync("/apps/gnome-osd",
                      dict(osd_halignment=dict(left=pango.ALIGN_LEFT,
                                               right=pango.ALIGN_RIGHT,
                                               center=pango.ALIGN_CENTER)))
    
    def osd_font_changed_cb(obj, key):
        print "'%s' changed" % key
    def osd_halignment_changed_cb(obj, key):
        print "'%s' changed" % key
        print "new value is ", getattr(prefs, key)
        
    prefs.connect("changed::osd_font", osd_font_changed_cb)
    prefs.connect("changed::osd_halignment", osd_halignment_changed_cb)

    print "osd font: ", prefs["osd_font"]
    prefs["osd_font"] = "Foo bar"
    print "osd font (after set): ", prefs['osd_font']

    print "osd_halignment is:", prefs['osd_halignment']
    print "setting osd_halignment = pango.ALIGN_LEFT"
    prefs['osd_halignment'] = pango.ALIGN_LEFT
    print "osd_halignment is:", prefs['osd_halignment']
    
    gtk.main()

